@@ -1,6 +1,6 @@
 /*
  * The initial "port" of dwm to curses was done by
- * (c) 2007-2009 Marc Andre Tanner <mat at brain-dump dot org>
+ * (c) 2007-2011 Marc Andre Tanner <mat at brain-dump dot org>
  *
  * It is highly inspired by the original X11 dwm and
  * reuses some code of it which is mostly
@@ -13,12 +13,17 @@
 #define _GNU_SOURCE
 #include <sys/stat.h>
 #include <sys/ioctl.h>
-#include <sys/fcntl.h>
+#ifdef _AIX
+# include <fcntl.h>
+#else
+# include <sys/fcntl.h>
+#endif
 #include <sys/wait.h>
 #include <sys/time.h>
 #include <sys/types.h>
-#include <ncurses.h>
+#include <curses.h>
 #include <stdio.h>
+#include <stdarg.h>
 #include <signal.h>
 #include <locale.h>
 #include <string.h>
@@ -28,7 +33,19 @@
 #ifdef __CYGWIN__
 # include <termios.h>
 #endif
-#include "madtty.h"
+#include "vt.h"
+
+#ifdef PDCURSES
+int ESCDELAY;
+#endif
+
+typedef struct {
+	double mfact;
+	int history;
+	int w;
+	int h;
+	bool need_resize;
+} Screen;
 
 typedef struct {
 	const char *symbol;
@@ -38,25 +55,30 @@ typedef struct {
 typedef struct Client Client;
 struct Client {
 	WINDOW *window;
-	madtty_t *term;
+	Vt *term;
 	const char *cmd;
 	char title[256];
 	uint8_t order;
 	pid_t pid;
 	int pty;
-#ifdef CONFIG_CMDFIFO
 	unsigned short int id;
-#endif
-	short int x;
-	short int y;
-	short int w;
-	short int h;
+	unsigned short int x;
+	unsigned short int y;
+	unsigned short int w;
+	unsigned short int h;
 	bool minimized;
 	bool died;
 	Client *next;
 	Client *prev;
 };
 
+typedef struct {
+	const char *title;
+	unsigned attrs;
+	short fg;
+	short bg;
+} ColorRule;
+
 #define ALT(k)      ((k) + (161 - 'a'))
 #ifndef CTRL
   #define CTRL(k)   ((k) & 0x1F)
@@ -78,23 +100,33 @@ typedef struct {
 	Action action;
 } Key;
 
-#ifdef CONFIG_MOUSE
 typedef struct {
 	mmask_t mask;
 	Action action;
 } Button;
-#endif
 
-#ifdef CONFIG_CMDFIFO
 typedef struct {
 	const char *name;
 	Action action;
 } Cmd;
-#endif
 
-#ifdef CONFIG_STATUSBAR
-enum { BarTop, BarBot, BarOff };
-#endif
+enum { BAR_TOP, BAR_BOTTOM, BAR_OFF };
+enum { ALIGN_LEFT, ALIGN_RIGHT };
+
+typedef struct {
+	int fd;
+	int pos;
+	unsigned short int h;
+	unsigned short int y;
+	char text[512];
+	const char *file;
+} StatusBar;
+
+typedef struct {
+	int fd;
+	const char *file;
+	unsigned short int id;
+} CmdFifo;
 
 #define countof(arr) (sizeof(arr) / sizeof((arr)[0]))
 #define sstrlen(str) (sizeof(str) - 1)
@@ -107,78 +139,54 @@ enum { BarTop, BarBot, BarOff };
 #endif
 
 /* commands for use by keybindings */
-static void quit(const char *args[]);
 static void create(const char *args[]);
-static void startup(const char *args[]);
 static void escapekey(const char *args[]);
-static void killclient(const char *args[]);
 static void focusn(const char *args[]);
 static void focusnext(const char *args[]);
 static void focusnextnm(const char *args[]);
 static void focusprev(const char *args[]);
 static void focusprevnm(const char *args[]);
+static void killclient(const char *args[]);
+static void lock(const char *key[]);
+static void quit(const char *args[]);
+static void redraw(const char *args[]);
+static void scrollback(const char *args[]);
+static void setlayout(const char *args[]);
+static void setmfact(const char *args[]);
+static void startup(const char *args[]);
+static void togglebar(const char *args[]);
 static void togglebell(const char *key[]);
 static void toggleminimize(const char *args[]);
-static void setmwfact(const char *args[]);
-static void setlayout(const char *args[]);
-static void scrollback(const char *args[]);
-static void redraw(const char *args[]);
-static void zoom(const char *args[]);
-static void lock(const char *key[]);
+static void togglemouse(const char *args[]);
 static void togglerunall(const char *args[]);
+static void zoom(const char *args[]);
 
-#ifdef CONFIG_STATUSBAR
-enum { ALIGN_LEFT, ALIGN_RIGHT };
-static void togglebar(const char *args[]);
-#endif
-
-#ifdef CONFIG_MOUSE
+/* commands for use by mouse bindings */
 static void mouse_focus(const char *args[]);
 static void mouse_fullscreen(const char *args[]);
 static void mouse_minimize(const char *args[]);
 static void mouse_zoom(const char *args[]);
-static void mouse_toggle();
-#endif
 
-static void clear_workspace();
-static void draw(Client *c);
-static void draw_all(bool border);
-static void draw_border(Client *c);
+/* functions and variables available to layouts via config.h */
 static void resize(Client *c, int x, int y, int w, int h);
-static void resize_screen();
-static void eprint(const char *errstr, ...);
-static bool isarrange(void (*func)());
-static void arrange();
-static void focus(Client *c);
-static void keypress(int code);
-
+extern Screen screen;
 static unsigned int waw, wah, wax, way;
 static Client *clients = NULL;
-extern double mwfact;
 
 #include "config.h"
 
+/* global variables */
+Screen screen = { MFACT, SCROLL_HISTORY };
 static Client *sel = NULL;
-double mwfact = MWFACT;
+static Client *msel = NULL;
+static bool mouse_events_enabled = ENABLE_MOUSE;
 static Layout *layout = layouts;
+static StatusBar bar = { -1, BAR_POS, 1 };
+static CmdFifo cmdfifo = { -1 };
 static const char *shell;
-static bool need_screen_resize;
-static int width, height, scroll_buf_size = SCROLL_BUF_SIZE;
 static bool running = true;
 static bool runinall = false;
 
-#ifdef CONFIG_MOUSE
-# include "mouse.c"
-#endif
-
-#ifdef CONFIG_CMDFIFO
-# include "cmdfifo.c"
-#endif
-
-#ifdef CONFIG_STATUSBAR
-# include "statusbar.c"
-#endif
-
 static void
 eprint(const char *errstr, ...) {
 	va_list ap;
@@ -196,6 +204,132 @@ error(const char *errstr, ...) {
 	exit(EXIT_FAILURE);
 }
 
+static bool
+isarrange(void (*func)()) {
+	return func == layout->arrange;
+}
+
+static void
+clear_workspace() {
+	for (unsigned int y = 0; y < wah; y++)
+		mvhline(way + y, 0, ' ', waw);
+	wnoutrefresh(stdscr);
+}
+
+static void
+drawbar() {
+	wchar_t wbuf[sizeof bar.text];
+	int w, maxwidth = screen.w - 2;
+	if (bar.pos == BAR_OFF || !bar.text[0])
+		return;
+	curs_set(0);
+	attrset(BAR_ATTR);
+	wcolor_set(stdscr, vt_color_get(BAR_FG, BAR_BG), NULL);
+	mvaddch(bar.y, 0, '[');
+	if (mbstowcs(wbuf, bar.text, sizeof bar.text) == (size_t)-1)
+		return;
+	if ((w = wcswidth(wbuf, maxwidth)) == -1)
+		return;
+	if (BAR_ALIGN == ALIGN_RIGHT) {
+		for (int i = 0; i + w < maxwidth; i++)
+			addch(' ');
+	}
+	addstr(bar.text);
+	if (BAR_ALIGN == ALIGN_LEFT) {
+		for (; w < maxwidth; w++)
+			addch(' ');
+	}
+	mvaddch(bar.y, screen.w - 1, ']');
+	attrset(NORMAL_ATTR);
+	if (sel)
+		curs_set(vt_cursor(sel->term));
+	refresh();
+}
+
+static void
+draw_border(Client *c) {
+	char *s, t = '\0';
+	int x, y, o;
+	if (sel == c) {
+		wattrset(c->window, SELECTED_ATTR);
+		wcolor_set(c->window, vt_color_get(SELECTED_FG, SELECTED_BG), NULL);
+	} else {
+		wattrset(c->window, NORMAL_ATTR);
+		wcolor_set(c->window, vt_color_get(NORMAL_FG, NORMAL_BG), NULL);
+	}
+	getyx(c->window, y, x);
+	curs_set(0);
+	mvwhline(c->window, 0, 0, ACS_HLINE, c->w);
+	o = c->w - (4 + sstrlen(TITLE) - 5  + sstrlen(SEPARATOR));
+	if (o < 0)
+		o = 0;
+	if ((size_t)o < sizeof(c->title)) {
+		t = *(s = &c->title[o]);
+		*s = '\0';
+	}
+	mvwprintw(c->window, 0, 2, TITLE,
+	          *c->title ? c->title : "",
+	          *c->title ? SEPARATOR : "",
+	          c->order);
+	if (t)
+		*s = t;
+	wmove(c->window, y, x);
+	if (!c->minimized)
+		curs_set(vt_cursor(c->term));
+}
+
+static void
+draw_content(Client *c) {
+	if (!c->minimized || isarrange(fullscreen)) {
+		vt_draw(c->term, c->window, 1, 0);
+		if (c != sel)
+			curs_set(0);
+	}
+}
+
+static void
+draw(Client *c) {
+	draw_content(c);
+	draw_border(c);
+	wrefresh(c->window);
+}
+
+static void
+draw_all(bool border) {
+	Client *c;
+	curs_set(0);
+	for (c = clients; c; c = c->next) {
+		redrawwin(c->window);
+		if (c == sel)
+			continue;
+		draw_content(c);
+		if (border)
+			draw_border(c);
+		wnoutrefresh(c->window);
+	}
+	/* as a last step the selected window is redrawn,
+	 * this has the effect that the cursor position is
+	 * accurate
+	 */
+	refresh();
+	if (sel) {
+		draw_content(sel);
+		if (border)
+			draw_border(sel);
+		wrefresh(sel->window);
+	}
+}
+
+static void
+arrange() {
+	clear_workspace();
+	attrset(NORMAL_ATTR);
+	color_set(vt_color_get(NORMAL_FG, NORMAL_BG), NULL);
+	layout->arrange();
+	wnoutrefresh(stdscr);
+	draw_all(true);
+}
+
 static void
 attach(Client *c) {
 	uint8_t order;
@@ -242,19 +376,6 @@ detach(Client *c) {
 	c->next = c->prev = NULL;
 }
 
-static void
-arrange() {
-	clear_workspace();
-	layout->arrange();
-	wnoutrefresh(stdscr);
-	draw_all(true);
-}
-
-static bool
-isarrange(void (*func)()) {
-	return func == layout->arrange;
-}
-
 static void
 focus(Client *c) {
 	Client *tmp = sel;
@@ -271,6 +392,343 @@ focus(Client *c) {
 	wrefresh(c->window);
 }
 
+static void
+applycolorrules(Vt *term, char *title) {
+	unsigned int i;
+	unsigned attrs = A_NORMAL;
+	short fg = -1, bg = -1;
+	const ColorRule *r;
+
+	for (i = 0; i < countof(colorrules); i++) {
+		r = &colorrules[i];
+		if (strstr(title, r->title)) {
+			attrs = r->attrs;
+			fg = r->fg;
+			bg = r->bg;
+			break;
+		}
+	}
+	vt_set_default_colors(term, attrs, fg, bg);
+}
+
+static int
+title_escape_seq_handler(Vt *term, char *es) {
+	Client *c;
+	unsigned int l;
+	if (es[0] != ']' || (es[1] && (es[1] < '0' || es[1] > '9')) || (es[2] && es[2] != ';'))
+		return VT_HANDLER_NOWAY;
+	if ((l = strlen(es)) < 3 || es[l - 1] != '\07')
+		return VT_HANDLER_NOTYET;
+	es[l - 1] = '\0';
+	c = (Client *)vt_get_data(term);
+	strncpy(c->title, es + 3, sizeof(c->title));
+	draw_border(c);
+	debug("window title: %s\n", c->title);
+	applycolorrules(term, c->title);
+	return VT_HANDLER_OK;
+}
+
+static void
+move_client(Client *c, int x, int y) {
+	if (c->x == x && c->y == y)
+		return;
+	debug("moving, x: %d y: %d\n", x, y);
+	if (mvwin(c->window, y, x) == ERR)
+		eprint("error moving, x: %d y: %d\n", x, y);
+	else {
+		c->x = x;
+		c->y = y;
+	}
+}
+
+static void
+resize_client(Client *c, int w, int h) {
+	if (c->w == w && c->h == h)
+		return;
+	debug("resizing, w: %d h: %d\n", w, h);
+	if (wresize(c->window, h, w) == ERR)
+		eprint("error resizing, w: %d h: %d\n", w, h);
+	else {
+		c->w = w;
+		c->h = h;
+	}
+	vt_resize(c->term, h - 1, w);
+}
+
+static void
+resize(Client *c, int x, int y, int w, int h) {
+	resize_client(c, w, h);
+	move_client(c, x, y);
+}
+
+static Client*
+get_client_by_pid(pid_t pid) {
+	Client *c;
+	for (c = clients; c; c = c->next) {
+		if (c->pid == pid)
+			return c;
+	}
+	return NULL;
+}
+
+static Client*
+get_client_by_coord(unsigned int x, unsigned int y) {
+	Client *c;
+	if (y < way || y >= wah)
+		return NULL;
+	if (isarrange(fullscreen))
+		return sel;
+	for (c = clients; c; c = c->next) {
+		if (x >= c->x && x < c->x + c->w && y >= c->y && y < c->y + c->h) {
+			debug("mouse event, x: %d y: %d client: %d\n", x, y, c->order);
+			return c;
+		}
+	}
+	return NULL;
+}
+
+static void
+sigchld_handler(int sig) {
+	int errsv = errno;
+	int status;
+	pid_t pid;
+	Client *c;
+
+	while ((pid = waitpid(-1, &status, WNOHANG)) != 0) {
+		if (pid == -1) {
+			if (errno == ECHILD) {
+				/* no more child processes */
+				break;
+			}
+			eprint("waitpid: %s\n", strerror(errno));
+			break;
+		}
+		debug("child with pid %d died\n", pid);
+		if ((c = get_client_by_pid(pid)))
+			c->died = true;
+	}
+
+	signal(SIGCHLD, sigchld_handler);
+
+	errno = errsv;
+}
+
+static void
+sigwinch_handler(int sig) {
+	signal(SIGWINCH, sigwinch_handler);
+	screen.need_resize = true;
+}
+
+static void
+sigterm_handler(int sig) {
+	running = false;
+}
+
+static void
+updatebarpos(void) {
+	bar.y = 0;
+	wax = 0;
+	way = 0;
+	wah = screen.h;
+	if (bar.fd == -1)
+		return;
+	if (bar.pos == BAR_TOP) {
+		wah -= bar.h;
+		way += bar.h;
+	} else if (bar.pos == BAR_BOTTOM) {
+		wah -= bar.h;
+		bar.y = wah;
+	}
+}
+
+static void
+resize_screen() {
+	struct winsize ws;
+
+	if (ioctl(0, TIOCGWINSZ, &ws) == -1)
+		return;
+
+	screen.w = ws.ws_col;
+	screen.h = ws.ws_row;
+
+	debug("resize_screen(), w: %d h: %d\n", screen.w, screen.h);
+
+	resizeterm(screen.h, screen.w);
+	wresize(stdscr, screen.h, screen.w);
+	wrefresh(curscr);
+	refresh();
+
+	waw = screen.w;
+	wah = screen.h;
+	updatebarpos();
+	drawbar();
+	arrange();
+}
+
+static bool
+is_modifier(unsigned int mod) {
+	unsigned int i;
+	for (i = 0; i < countof(keys); i++) {
+		if (keys[i].mod == mod)
+			return true;
+	}
+	return false;
+}
+
+static Key*
+keybinding(unsigned int mod, unsigned int code) {
+	unsigned int i;
+	for (i = 0; i < countof(keys); i++) {
+		if (keys[i].mod == mod && keys[i].code == code)
+			return &keys[i];
+	}
+	return NULL;
+}
+
+static void
+keypress(int code) {
+	Client *c;
+	unsigned int len = 1;
+	char buf[8] = { '\e' };
+
+	if (code == '\e') {
+		/* pass characters following escape to the underlying app */
+		nodelay(stdscr, TRUE);
+		for (int t; len < sizeof(buf) && (t = getch()) != ERR; len++)
+			buf[len] = t;
+		nodelay(stdscr, FALSE);
+	}
+
+	for (c = runinall ? clients : sel; c; c = c->next) {
+		if (!c->minimized || isarrange(fullscreen)) {
+			if (code == '\e')
+				vt_write(c->term, buf, len);
+			else
+				vt_keypress(c->term, code);
+		}
+		if (!runinall)
+			break;
+	}
+}
+
+static void
+mouse_setup() {
+#ifdef CONFIG_MOUSE
+	mmask_t mask = 0;
+
+	if (mouse_events_enabled) {
+		mask = BUTTON1_CLICKED | BUTTON2_CLICKED;
+		for (unsigned int i = 0; i < countof(buttons); i++)
+			mask |= buttons[i].mask;
+	}
+	mousemask(mask, NULL);
+#endif /* CONFIG_MOUSE */
+}
+
+static void
+setup() {
+	if (!(shell = getenv("SHELL")))
+		shell = "/bin/sh";
+	setlocale(LC_CTYPE, "");
+	initscr();
+	start_color();
+	noecho();
+	keypad(stdscr, TRUE);
+	mouse_setup();
+	raw();
+	vt_init();
+	getmaxyx(stdscr, screen.h, screen.w);
+	resize_screen();
+	signal(SIGWINCH, sigwinch_handler);
+	signal(SIGCHLD, sigchld_handler);
+	signal(SIGTERM, sigterm_handler);
+}
+
+static void
+destroy(Client *c) {
+	if (sel == c)
+		focusnextnm(NULL);
+	detach(c);
+	if (sel == c) {
+		if (clients) {
+			focus(clients);
+			toggleminimize(NULL);
+		} else
+			sel = NULL;
+	}
+	werase(c->window);
+	wrefresh(c->window);
+	vt_destroy(c->term);
+	delwin(c->window);
+	if (!clients && countof(actions)) {
+		if (!strcmp(c->cmd, shell))
+			quit(NULL);
+		else
+			create(NULL);
+	}
+	free(c);
+	arrange();
+}
+
+static void
+cleanup() {
+	endwin();
+	if (bar.fd > 0)
+		close(bar.fd);
+	if (bar.file)
+		unlink(bar.file);
+	if (cmdfifo.fd > 0)
+		close(cmdfifo.fd);
+	if (cmdfifo.file)
+		unlink(cmdfifo.file);
+}
+
+/* commands for use by keybindings */
+static void
+create(const char *args[]) {
+	Client *c = calloc(1, sizeof(Client));
+	if (!c)
+		return;
+	const char *cmd = (args && args[0]) ? args[0] : shell;
+	const char *pargs[] = { "/bin/sh", "-c", cmd, NULL };
+	c->id = ++cmdfifo.id;
+	char buf[8];
+	snprintf(buf, sizeof buf, "%d", c->id);
+	const char *env[] = {
+		"DVTM", VERSION,
+		"DVTM_WINDOW_ID", buf,
+		NULL
+	};
+
+	c->window = newwin(wah, waw, way, wax);
+	c->term = vt_create(screen.h - 1, screen.w, screen.history);
+	c->cmd = cmd;
+	if (args && args[1])
+		strncpy(c->title, args[1], sizeof(c->title));
+	c->pid = vt_forkpty(c->term, "/bin/sh", pargs, env, &c->pty);
+	vt_set_data(c->term, c);
+	vt_set_handler(c->term, title_escape_seq_handler);
+	c->w = screen.w;
+	c->h = screen.h;
+	c->x = wax;
+	c->y = way;
+	c->order = 0;
+	c->minimized = false;
+	debug("client with pid %d forked\n", c->pid);
+	attach(c);
+	focus(c);
+	arrange();
+}
+
+static void
+escapekey(const char *args[]) {
+	int key;
+	if ((key = getch()) >= 0) {
+		debug("escaping key `%c'\n", key);
+		keypress(CTRL(key));
+	}
+}
+
 static void
 focusn(const char *args[]) {
 	Client *c;
@@ -343,25 +801,129 @@ focusprevnm(const char *args[]) {
 }
 
 static void
-zoom(const char *args[]) {
-	Client *c;
-
+killclient(const char *args[]) {
 	if (!sel)
 		return;
-	if ((c = sel) == clients)
-		if (!(c = c->next))
-			return;
-	detach(c);
-	attach(c);
-	focus(c);
-	if (c->minimized)
-		toggleminimize(NULL);
+	debug("killing client with pid: %d\n", sel->pid);
+	kill(-sel->pid, SIGKILL);
+}
+
+static void
+lock(const char *args[]) {
+	size_t len = 0, i = 0;
+	char buf[16], *pass = buf;
+	int c;
+
+	erase();
+	curs_set(0);
+
+	if (args && args[0]) {
+		len = strlen(args[0]);
+		pass = (char *)args[0];
+	} else {
+		mvprintw(LINES / 2, COLS / 2 - 7, "Enter password");
+		while (len < sizeof buf && (c = getch()) != '\n')
+			if (c != ERR)
+				buf[len++] = c;
+	}
+
+	mvprintw(LINES / 2, COLS / 2 - 7, "Screen locked!");
+
+	while (i != len) {
+		for(i = 0; i < len; i++) {
+			if (getch() != pass[i])
+				break;
+		}
+	}
+
 	arrange();
 }
 
+static void
+quit(const char *args[]) {
+	cleanup();
+	exit(EXIT_SUCCESS);
+}
+
+static void
+redraw(const char *args[]) {
+	wrefresh(curscr);
+	resize_screen();
+	draw_all(true);
+}
+
+static void
+scrollback(const char *args[]) {
+	if (!sel) return;
+
+	if (!args[0] || atoi(args[0]) < 0)
+		vt_scroll(sel->term, -sel->h/2);
+	else
+		vt_scroll(sel->term,  sel->h/2);
+
+	draw(sel);
+}
+
+static void
+setlayout(const char *args[]) {
+	unsigned int i;
+
+	if (!args || !args[0]) {
+		if (++layout == &layouts[countof(layouts)])
+			layout = &layouts[0];
+	} else {
+		for (i = 0; i < countof(layouts); i++)
+			if (!strcmp(args[0], layouts[i].symbol))
+				break;
+		if (i == countof(layouts))
+			return;
+		layout = &layouts[i];
+	}
+	arrange();
+}
+
+static void
+setmfact(const char *args[]) {
+	double delta;
+
+	if (isarrange(fullscreen) || isarrange(grid))
+		return;
+	/* arg handling, manipulate mfact */
+	if (args[0] == NULL)
+		screen.mfact = MFACT;
+	else if (1 == sscanf(args[0], "%lf", &delta)) {
+		if (args[0][0] == '+' || args[0][0] == '-')
+			screen.mfact += delta;
+		else
+			screen.mfact = delta;
+		if (screen.mfact < 0.1)
+			screen.mfact = 0.1;
+		else if (screen.mfact > 0.9)
+			screen.mfact = 0.9;
+	}
+	arrange();
+}
+
+static void
+startup(const char *args[]) {
+	for (unsigned int i = 0; i < countof(actions); i++)
+		actions[i].cmd(actions[i].args);
+}
+
+static void
+togglebar(const char *args[]) {
+	if (bar.pos == BAR_OFF)
+		bar.pos = (BAR_POS == BAR_OFF) ? BAR_TOP : BAR_POS;
+	else
+		bar.pos = BAR_OFF;
+	updatebarpos();
+	arrange();
+	drawbar();
+}
+
 static void
 togglebell(const char *args[]) {
-	madtty_togglebell(sel->term);
+	vt_togglebell(sel->term);
 }
 
 static void
@@ -397,7 +959,7 @@ toggleminimize(const char *args[]) {
 		for (c = clients; c && c->next && !c->next->minimized; c = c->next);
 		attachafter(m, c);
 	} else { /* window is no longer minimized, move it to the master area */
-		madtty_dirty(m->term);
+		vt_dirty(m->term);
 		detach(m);
 		attach(m);
 	}
@@ -405,194 +967,9 @@ toggleminimize(const char *args[]) {
 }
 
 static void
-setlayout(const char *args[]) {
-	unsigned int i;
-
-	if (!args || !args[0]) {
-		if (++layout == &layouts[countof(layouts)])
-			layout = &layouts[0];
-	} else {
-		for (i = 0; i < countof(layouts); i++)
-			if (!strcmp(args[0], layouts[i].symbol))
-				break;
-		if (i == countof(layouts))
-			return;
-		layout = &layouts[i];
-	}
-	arrange();
-}
-
-static void
-setmwfact(const char *args[]) {
-	double delta;
-
-	if (isarrange(fullscreen) || isarrange(grid))
-		return;
-	/* arg handling, manipulate mwfact */
-	if (args[0] == NULL)
-		mwfact = MWFACT;
-	else if (1 == sscanf(args[0], "%lf", &delta)) {
-		if (args[0][0] == '+' || args[0][0] == '-')
-			mwfact += delta;
-		else
-			mwfact = delta;
-		if (mwfact < 0.1)
-			mwfact = 0.1;
-		else if (mwfact > 0.9)
-			mwfact = 0.9;
-	}
-	arrange();
-}
-
-static void
-scrollback(const char *args[]) {
-	if (!sel) return;
-
-	if (!args[0] || atoi(args[0]) < 0)
-		madtty_scroll(sel->term, -sel->h/2);
-	else
-		madtty_scroll(sel->term,  sel->h/2);
-
-	draw(sel);
-}
-
-static void
-redraw(const char *args[]) {
-	wrefresh(curscr);
-	resize_screen();
-	draw_all(true);
-}
-
-static void
-draw_border(Client *c) {
-	char *s, t = '\0';
-	int x, y, o;
-	if (sel == c) {
-		wattrset(c->window, SELECTED_ATTR);
-		madtty_color_set(c->window, SELECTED_FG, SELECTED_BG);
-	} else {
-		wattrset(c->window, NORMAL_ATTR);
-		madtty_color_set(c->window, NORMAL_FG, NORMAL_BG);
-	}
-	getyx(c->window, y, x);
-	curs_set(0);
-	mvwhline(c->window, 0, 0, ACS_HLINE, c->w);
-	o = c->w - (4 + sstrlen(TITLE) - 5  + sstrlen(SEPARATOR));
-	if (o < 0)
-		o = 0;
-	if (o < sizeof(c->title)) {
-		t = *(s = &c->title[o]);
-		*s = '\0';
-	}
-	mvwprintw(c->window, 0, 2, TITLE,
-	          *c->title ? c->title : "",
-	          *c->title ? SEPARATOR : "",
-	          c->order);
-	if (t)
-		*s = t;
-	wmove(c->window, y, x);
-	if (!c->minimized)
-		curs_set(madtty_cursor(c->term));
-}
-
-static void
-draw_content(Client *c) {
-	if (!c->minimized || isarrange(fullscreen)) {
-		madtty_draw(c->term, c->window, 1, 0);
-		if (c != sel)
-			curs_set(0);
-	}
-}
-
-static void
-draw(Client *c) {
-	draw_content(c);
-	draw_border(c);
-	wrefresh(c->window);
-}
-
-static void
-clear_workspace() {
-	unsigned int y;
-	for (y = 0; y < wah; y++)
-		mvhline(way + y, 0, ' ', waw);
-	wnoutrefresh(stdscr);
-}
-
-static void
-draw_all(bool border) {
-	Client *c;
-	curs_set(0);
-	for (c = clients; c; c = c->next) {
-		redrawwin(c->window);
-		if (c == sel)
-			continue;
-		draw_content(c);
-		if (border)
-			draw_border(c);
-		wnoutrefresh(c->window);
-	}
-	/* as a last step the selected window is redrawn,
-	 * this has the effect that the cursor position is
-	 * accurate
-	 */
-	refresh();
-	if (sel) {
-		draw_content(sel);
-		if (border)
-			draw_border(sel);
-		wrefresh(sel->window);
-	}
-}
-
-static void
-escapekey(const char *args[]) {
-	int key;
-	if ((key = getch()) >= 0) {
-		debug("escaping key `%c'\n", key);
-		keypress(CTRL(key));
-	}
-}
-
-/*
- * Lock the screen until the correct password is entered.
- * The password can either be specified in config.h which is
- * not recommended because `strings dvtm` will contain it. If
- * no password is specified in the configuration file it is read
- * from the keyboard before the screen is locked.
- *
- * NOTE: this function doesn't handle the input from clients. All
- *       foreground operations are temporarily suspended since the
- *       function doesn't return.
- */
-static void
-lock(const char *args[]) {
-	size_t len = 0, i = 0;
-	char buf[16], *pass = buf, c;
-
-	erase();
-	curs_set(0);
-
-	if (args && args[0]) {
-		len = strlen(args[0]);
-		pass = (char *)args[0];
-	} else {
-		mvprintw(LINES / 2, COLS / 2 - 7, "Enter password");
-		while (len < sizeof buf && (c = getch()) != '\n')
-			if (c != ERR)
-				buf[len++] = c;
-	}
-
-	mvprintw(LINES / 2, COLS / 2 - 7, "Screen locked!");
-
-	while (i != len) {
-		for(i = 0; i < len; i++) {
-			if (getch() != pass[i])
-				break;
-		}
-	}
-
-	arrange();
+togglemouse(const char *args[]) {
+	mouse_events_enabled = !mouse_events_enabled;
+	mouse_setup();
 }
 
 static void
@@ -601,305 +978,236 @@ togglerunall(const char *args[]) {
 }
 
 static void
-killclient(const char *args[]) {
+zoom(const char *args[]) {
+	Client *c;
+
 	if (!sel)
 		return;
-	debug("killing client with pid: %d\n", sel->pid);
-	kill(-sel->pid, SIGKILL);
+	if ((c = sel) == clients)
+		if (!(c = c->next))
+			return;
+	detach(c);
+	attach(c);
+	focus(c);
+	if (c->minimized)
+		toggleminimize(NULL);
+	arrange();
+}
+
+/* commands for use by mouse bindings */
+static void
+mouse_focus(const char *args[]) {
+	focus(msel);
+	if (msel->minimized)
+		toggleminimize(NULL);
+}
+
+static void
+mouse_fullscreen(const char *args[]) {
+	mouse_focus(NULL);
+	if (isarrange(fullscreen))
+		setlayout(NULL);
+	else
+		setlayout(args);
+}
+
+static void
+mouse_minimize(const char *args[]) {
+	focus(msel);
+	toggleminimize(NULL);
+}
+
+static void
+mouse_zoom(const char *args[]) {
+	focus(msel);
+	zoom(NULL);
+}
+
+static Cmd *
+get_cmd_by_name(const char *name) {
+	for (unsigned int i = 0; i < countof(commands); i++) {
+		if (!strcmp(name, commands[i].name))
+			return &commands[i];
+	}
+	return NULL;
+}
+
+static void
+handle_cmdfifo() {
+	int r;
+	char *p, *s, cmdbuf[512], c;
+	Cmd *cmd;
+	switch (r = read(cmdfifo.fd, cmdbuf, sizeof cmdbuf - 1)) {
+	case -1:
+	case 0:
+		cmdfifo.fd = -1;
+		break;
+	default:
+		cmdbuf[r] = '\0';
+		p = cmdbuf;
+		while (*p) {
+			/* find the command name */
+			for (; *p == ' ' || *p == '\n'; p++);
+			for (s = p; *p && *p != ' ' && *p != '\n'; p++);
+			if ((c = *p))
+				*p++ = '\0';
+			if (*s && (cmd = get_cmd_by_name(s)) != NULL) {
+				bool quote = false;
+				int argc = 0;
+				/* XXX: initializer assumes MAX_ARGS == 2 use a initialization loop? */
+				const char *args[MAX_ARGS] = { NULL, NULL}, *arg;
+				/* if arguments were specified in config.h ignore the one given via
+				 * the named pipe and thus skip everything until we find a new line
+				 */
+				if (cmd->action.args[0] || c == '\n') {
+					debug("execute %s", s);
+					cmd->action.cmd(cmd->action.args);
+					while (*p && *p != '\n')
+						p++;
+					continue;
+				}
+				/* no arguments were given in config.h so we parse the command line */
+				while (*p == ' ')
+					p++;
+				arg = p;
+				for (; (c = *p); p++) {
+					switch (*p) {
+					case '\\':
+						/* remove the escape character '\\' move every
+						 * following character to the left by one position
+						 */
+						switch (*(++p)) {
+							case '\\':
+							case '\'':
+							case '\"': {
+								char *t = p;
+								for (;;) {
+									*(t - 1) = *t;
+									if (*t++ == '\0')
+										break;
+								}
+								p -= 2;
+							}
+						}
+						break;
+					case '\'':
+					case '\"':
+						quote = !quote;
+						break;
+					case ' ':
+						if (!quote) {
+					case '\n':
+							/* remove trailing quote if there is one */
+							if (*(p - 1) == '\'' || *(p - 1) == '\"')
+								*(p - 1) = '\0';
+							*p++ = '\0';
+							/* remove leading quote if there is one */
+							if (*arg == '\'' || *arg == '\"')
+								arg++;
+							if (argc < MAX_ARGS)
+								args[argc++] = arg;
+
+							while (*p == ' ')
+								++p;
+							arg = p;
+						}
+						break;
+					}
+
+					if (c == '\n' || *p == '\n') {
+						debug("execute %s", s);
+						for(int i = 0; i < argc; i++)
+							debug(" %s", args[i]);
+						debug("\n");
+						cmd->action.cmd(args);
+						break;
+					}
+				}
+			}
+		}
+	}
+}
+
+static void
+handle_mouse() {
+#ifdef CONFIG_MOUSE
+	MEVENT event;
+	unsigned int i;
+	if (getmouse(&event) != OK)
+		return;
+	msel = get_client_by_coord(event.x, event.y);
+
+	if (!msel)
+		return;
+
+	debug("mouse x:%d y:%d cx:%d cy:%d mask:%d\n", event.x, event.y, event.x - msel->x, event.y - msel->y, event.bstate);
+
+	vt_mouse(msel->term, event.x - msel->x, event.y - msel->y, event.bstate);
+
+	for (i = 0; i < countof(buttons); i++) {
+		if (event.bstate & buttons[i].mask)
+			buttons[i].action.cmd(buttons[i].action.args);
+	}
+
+	msel = NULL;
+#endif /* CONFIG_MOUSE */
+}
+
+static void
+handle_statusbar() {
+	char *p;
+	int r;
+	switch (r = read(bar.fd, bar.text, sizeof bar.text - 1)) {
+		case -1:
+			strncpy(bar.text, strerror(errno), sizeof bar.text - 1);
+			bar.text[sizeof bar.text - 1] = '\0';
+			bar.fd = -1;
+			break;
+		case 0:
+			bar.fd = -1;
+			break;
+		default:
+			bar.text[r] = '\0'; p = bar.text + strlen(bar.text) - 1;
+			for (; p >= bar.text && *p == '\n'; *p-- = '\0');
+			for (; p >= bar.text && *p != '\n'; --p);
+			if (p > bar.text)
+				strncpy(bar.text, p + 1, sizeof bar.text);
+			drawbar();
+	}
 }
 
 static int
-title_escape_seq_handler(madtty_t *term, char *es) {
-	Client *c;
-	unsigned int l;
-	if (es[0] != ']' || (es[1] && (es[1] < '0' || es[1] > '9')) || (es[2] && es[2] != ';'))
-		return MADTTY_HANDLER_NOWAY;
-	if ((l = strlen(es)) < 3 || es[l - 1] != '\07')
-		return MADTTY_HANDLER_NOTYET;
-	es[l - 1] = '\0';
-	c = (Client *)madtty_get_data(term);
-	strncpy(c->title, es + 3, sizeof(c->title));
-	draw_border(c);
-	debug("window title: %s\n", c->title);
-	return MADTTY_HANDLER_OK;
-}
+open_or_create_fifo(const char *name, const char **name_created) {
+	struct stat info;
+	int fd;
 
-static void
-create(const char *args[]) {
-	Client *c = calloc(sizeof(Client), 1);
-	if (!c)
-		return;
-	const char *cmd = (args && args[0]) ? args[0] : shell;
-	const char *pargs[] = { "/bin/sh", "-c", cmd, NULL };
-#ifdef CONFIG_CMDFIFO
-	c->id = ++client_id;
-	char buf[8];
-	snprintf(buf, sizeof buf, "%d", c->id);
-#endif
-	const char *env[] = {
-		"DVTM", VERSION,
-#ifdef CONFIG_CMDFIFO
-		"DVTM_WINDOW_ID", buf,
-#endif
-		NULL
-	};
-
-	c->window = newwin(wah, waw, way, wax);
-	c->term = madtty_create(height - 1, width, scroll_buf_size);
-	c->cmd = cmd;
-	if (args && args[1])
-		strncpy(c->title, args[1], sizeof(c->title));
-	c->pid = madtty_forkpty(c->term, "/bin/sh", pargs, env, &c->pty);
-	madtty_set_data(c->term, c);
-	madtty_set_handler(c->term, title_escape_seq_handler);
-	c->w = width;
-	c->h = height;
-	c->x = wax;
-	c->y = way;
-	c->order = 0;
-	c->minimized = false;
-	debug("client with pid %d forked\n", c->pid);
-	attach(c);
-	focus(c);
-	arrange();
-}
-
-static void
-destroy(Client *c) {
-	if (sel == c)
-		focusnextnm(NULL);
-	detach(c);
-	if (sel == c) {
-		if (clients) {
-			focus(clients);
-			toggleminimize(NULL);
-		} else
-			sel = NULL;
-	}
-	werase(c->window);
-	wrefresh(c->window);
-	madtty_destroy(c->term);
-	delwin(c->window);
-	if (!clients && countof(actions)) {
-		if (!strcmp(c->cmd, shell))
-			quit(NULL);
-		else
-			create(NULL);
-	}
-	free(c);
-	arrange();
-}
-
-static void
-move_client(Client *c, int x, int y) {
-	if (c->x == x && c->y == y)
-		return;
-	debug("moving, x: %d y: %d\n", x, y);
-	if (mvwin(c->window, y, x) == ERR)
-		eprint("error moving, x: %d y: %d\n", x, y);
-	else {
-		c->x = x;
-		c->y = y;
-	}
-}
-
-static void
-resize_client(Client *c, int w, int h) {
-	if (c->w == w && c->h == h)
-		return;
-	debug("resizing, w: %d h: %d\n", w, h);
-	if (wresize(c->window, h, w) == ERR)
-		eprint("error resizing, w: %d h: %d\n", w, h);
-	else {
-		c->w = w;
-		c->h = h;
-	}
-	madtty_resize(c->term, h - 1, w);
-}
-
-static void
-resize(Client *c, int x, int y, int w, int h) {
-	resize_client(c, w, h);
-	move_client(c, x, y);
-}
-
-static bool
-is_modifier(unsigned int mod) {
-	unsigned int i;
-	for (i = 0; i < countof(keys); i++) {
-		if (keys[i].mod == mod)
-			return true;
-	}
-	return false;
-}
-
-static Key*
-keybinding(unsigned int mod, unsigned int code) {
-	unsigned int i;
-	for (i = 0; i < countof(keys); i++) {
-		if (keys[i].mod == mod && keys[i].code == code)
-			return &keys[i];
-	}
-	return NULL;
-}
-
-static Client*
-get_client_by_pid(pid_t pid) {
-	Client *c;
-	for (c = clients; c; c = c->next) {
-		if (c->pid == pid)
-			return c;
-	}
-	return NULL;
-}
-
-static void
-sigchld_handler(int sig) {
-	int errsv = errno;
-	int status;
-	pid_t pid;
-	Client *c;
-
-	signal(SIGCHLD, sigchld_handler);
-
-	while ((pid = waitpid(-1, &status, WNOHANG)) != 0) {
-		if (pid == -1) {
-			if (errno == ECHILD) {
-				/* no more child processes */
-				break;
+	do {
+		if ((fd = open(name, O_RDWR|O_NONBLOCK)) == -1) {
+			if (errno == ENOENT && !mkfifo(name, S_IRUSR|S_IWUSR)) {
+				*name_created = name;
+				continue;
 			}
-			eprint("waitpid: %s\n", strerror(errno));
-			break;
+			error("%s\n", strerror(errno));
 		}
-		debug("child with pid %d died\n", pid);
-		if ((c = get_client_by_pid(pid)))
-			c->died = true;
-	}
+	} while (fd == -1);
 
-	errno = errsv;
-}
-
-static void
-sigwinch_handler(int sig) {
-	signal(SIGWINCH, sigwinch_handler);
-	need_screen_resize = true;
-}
-
-static void
-sigterm_handler(int sig) {
-	running = false;
-}
-
-static void
-resize_screen() {
-	struct winsize ws;
-
-	if (ioctl(0, TIOCGWINSZ, &ws) == -1)
-		return;
-
-	width = ws.ws_col;
-	height = ws.ws_row;
-
-	debug("resize_screen(), w: %d h: %d\n", width, height);
-
-#if defined(__OpenBSD__) || defined(__NetBSD__)
-	resizeterm(height, width);
-#else
-	resize_term(height, width);
-#endif
-	wresize(stdscr, height, width);
-	wrefresh(curscr);
-	refresh();
-
-	waw = width;
-	wah = height;
-#ifdef CONFIG_STATUSBAR
-	updatebarpos();
-	drawbar();
-#endif
-	arrange();
-}
-
-static void
-startup(const char *args[]) {
-	for (int i = 0; i < countof(actions); i++)
-		actions[i].cmd(actions[i].args);
-}
-
-static void
-setup() {
-	if (!(shell = getenv("SHELL")))
-		shell = "/bin/sh";
-	setlocale(LC_CTYPE, "");
-	initscr();
-	start_color();
-	noecho();
-	keypad(stdscr, TRUE);
-#ifdef CONFIG_MOUSE
-	mouse_setup();
-#endif
-	raw();
-	madtty_init_colors();
-	madtty_init_vt100_graphics();
-	getmaxyx(stdscr, height, width);
-	resize_screen();
-	signal(SIGWINCH, sigwinch_handler);
-	signal(SIGCHLD, sigchld_handler);
-	signal(SIGTERM, sigterm_handler);
-}
-
-static void
-cleanup() {
-	endwin();
-#ifdef CONFIG_STATUSBAR
-	if (statusfd > 0)
-		close(statusfd);
-#endif
-#ifdef CONFIG_CMDFIFO
-	if (cmdfd > 0)
-		close(cmdfd);
-	if (cmdpath)
-		unlink(cmdpath);
-#endif
-}
-
-static void
-quit(const char *args[]) {
-	cleanup();
-	exit(EXIT_SUCCESS);
+	if (fstat(fd, &info) == -1)
+		error("%s\n", strerror(errno));
+	if (!S_ISFIFO(info.st_mode))
+		error("%s is not a named pipe\n", name);
+	return fd;
 }
 
 static void
 usage() {
 	cleanup();
 	eprint("usage: dvtm [-v] [-m mod] [-d escdelay] [-h n] "
-#ifdef CONFIG_STATUSBAR
 		"[-s status-fifo] "
-#endif
-#ifdef CONFIG_CMDFIFO
 		"[-c cmd-fifo] "
-#endif
 		"[cmd...]\n");
 	exit(EXIT_FAILURE);
 }
 
-static int
-open_or_create_fifo(const char *name) {
-	struct stat info;
-	int fd;
-open:
-	if ((fd = open(name, O_RDWR|O_NONBLOCK)) == -1) {
-		if (errno == ENOENT && !mkfifo(name, S_IRUSR|S_IWUSR))
-			goto open;
-		error("%s\n", strerror(errno));
-	}
-	if (fstat(fd, &info) == -1)
-		error("%s\n", strerror(errno));
-	if (!S_ISFIFO(info.st_mode))
-		error("%s is not a named pipe\n", name);
-	return fd;
-}
-
 static bool
 parse_args(int argc, char *argv[]) {
 	int arg;
@@ -921,13 +1229,13 @@ parse_args(int argc, char *argv[]) {
 			usage();
 		switch (argv[arg][1]) {
 			case 'v':
-				puts("dvtm-"VERSION" (c) 2007-2009 Marc Andre Tanner");
+				puts("dvtm-"VERSION" (c) 2007-2011 Marc Andre Tanner");
 				exit(EXIT_SUCCESS);
 			case 'm': {
 				char *mod = argv[++arg];
 				if (mod[0] == '^' && mod[1])
 					*mod = CTRL(mod[1]);
-				for (int i = 0; i < countof(keys); i++)
+				for (unsigned int i = 0; i < countof(keys); i++)
 					keys[i].mod = *mod;
 				break;
 			}
@@ -939,22 +1247,20 @@ parse_args(int argc, char *argv[]) {
 					ESCDELAY = 1000;
 				break;
 			case 'h':
-				scroll_buf_size = atoi(argv[++arg]);
+				screen.history = atoi(argv[++arg]);
 				break;
-#ifdef CONFIG_STATUSBAR
 			case 's':
-				statusfd = open_or_create_fifo(argv[++arg]);
+				bar.fd = open_or_create_fifo(argv[++arg], &bar.file);
 				updatebarpos();
 				break;
-#endif
-#ifdef CONFIG_CMDFIFO
-			case 'c':
-				cmdfd = open_or_create_fifo(argv[++arg]);
-				if (!(cmdpath = get_realpath(argv[arg])))
+			case 'c': {
+				const char *fifo;
+				cmdfifo.fd = open_or_create_fifo(argv[++arg], &cmdfifo.file);
+				if (!(fifo = realpath(argv[arg], NULL)))
 					error("%s\n", strerror(errno));
-				setenv("DVTM_CMD_FIFO", cmdpath, 1);
+				setenv("DVTM_CMD_FIFO", fifo, 1);
 				break;
-#endif
+			}
 			default:
 				usage();
 		}
@@ -962,33 +1268,6 @@ parse_args(int argc, char *argv[]) {
 	return init;
 }
 
-void
-keypress(int code) {
-	Client *c;
-	int len = 1;
-	char buf[8] = { '\e' };
-
-	if (code == '\e') {
-		/* pass characters following escape to the underlying app */
-		nodelay(stdscr, TRUE);
-		while (len < sizeof(buf) - 1 && (buf[len] = getch()) != ERR)
-			len++;
-		buf[len] = '\0';
-		nodelay(stdscr, FALSE);
-	}
-
-	for (c = runinall ? clients : sel; c; c = c->next) {
-		if (!c->minimized || isarrange(fullscreen)) {
-			if (code == '\e')
-				madtty_keypress_sequence(c->term, buf);
-			else
-				madtty_keypress(c->term, code);
-		}
-		if (!runinall)
-			break;
-	}
-}
-
 int
 main(int argc, char *argv[]) {
 	if (!parse_args(argc, argv)) {
@@ -1001,26 +1280,24 @@ main(int argc, char *argv[]) {
 		int r, nfds = 0;
 		fd_set rd;
 
-		if (need_screen_resize) {
+		if (screen.need_resize) {
 			resize_screen();
-			need_screen_resize = false;
+			screen.need_resize = false;
 		}
 
 		FD_ZERO(&rd);
 		FD_SET(STDIN_FILENO, &rd);
 
-#ifdef CONFIG_CMDFIFO
-		if (cmdfd != -1) {
-			FD_SET(cmdfd, &rd);
-			nfds = cmdfd;
+		if (cmdfifo.fd != -1) {
+			FD_SET(cmdfifo.fd, &rd);
+			nfds = cmdfifo.fd;
 		}
-#endif
-#ifdef CONFIG_STATUSBAR
-		if (statusfd != -1) {
-			FD_SET(statusfd, &rd);
-			nfds = max(nfds, statusfd);
+
+		if (bar.fd != -1) {
+			FD_SET(bar.fd, &rd);
+			nfds = max(nfds, bar.fd);
 		}
-#endif
+
 		for (c = clients; c; ) {
 			if (c->died) {
 				t = c->next;
@@ -1046,12 +1323,9 @@ main(int argc, char *argv[]) {
 			int code = getch();
 			Key *key;
 			if (code >= 0) {
-#ifdef CONFIG_MOUSE
 				if (code == KEY_MOUSE) {
 					handle_mouse();
-				} else
-#endif /* CONFIG_MOUSE */
-				if (is_modifier(code)) {
+				} else if (is_modifier(code)) {
 					int mod = code;
 					code = getch();
 					if (code >= 0) {
@@ -1070,18 +1344,15 @@ main(int argc, char *argv[]) {
 				continue;
 		}
 
-#ifdef CONFIG_CMDFIFO
-		if (cmdfd != -1 && FD_ISSET(cmdfd, &rd))
+		if (cmdfifo.fd != -1 && FD_ISSET(cmdfifo.fd, &rd))
 			handle_cmdfifo();
-#endif
-#ifdef CONFIG_STATUSBAR
-		if (statusfd != -1 && FD_ISSET(statusfd, &rd))
+
+		if (bar.fd != -1 && FD_ISSET(bar.fd, &rd))
 			handle_statusbar();
-#endif
 
 		for (c = clients; c; ) {
 			if (FD_ISSET(c->pty, &rd)) {
-				if (madtty_process(c->term) < 0 && errno == EIO) {
+				if (vt_process(c->term) < 0 && errno == EIO) {
 					/* client probably terminated */
 					t = c->next;
 					destroy(c);
